class Store:
    def __init__(self, store_url=None):
        pass

    def load_entries(self):
        return {}

    def save_entries(self, entries):
        pass

    def has_locked(self):
        return True

    def acquire_lock(self, lock_ttl=None):
        return True

    def renew_lock(self):
        pass

    def release_lock(self):
        pass



from urllib.parse import urlparse

from celery.beat import Scheduler as BaseScheduler
from celery.exceptions import ImproperlyConfigured
from celery.utils.log import get_logger

from .utils import import_string

logger = get_logger(__name__)


def get_store(app):
    """
    Get store object fro celery application.

    :param app: celery application
    :return: store
    """
    store_classes = getattr(app.conf, 'CELERY_BEAT_STORE_CLASSES', {
        'dummy': 'beatx.store.dummy.Store',
        'redis': 'beatx.store.redis.Store',
    })
    store_url = getattr(app.conf, 'CELERY_BEAT_STORE', 'dummy://')

    scheme = urlparse(store_url).scheme

    try:
        store_class = store_classes[scheme]
    except KeyError:
        raise ImproperlyConfigured(
            '"%(used)s" store scheme is unsupported. Available stores schemes: %(available)s' % {
                'used': scheme,
                'available': ', '.join('"%s"' % x for x in store_classes.keys())
            }
        )

    if not isinstance(store_class, type):
        store_class = import_string(store_class)

    return store_class(store_url)


class Scheduler(BaseScheduler):
    """
    Celery scheduler.
    
    Simple celery scheduler which use store class to load/save schedule.
    """

    def __init__(self, app, *args, **kwargs):
        self.store = get_store(app)

        super().__init__(app, *args, **kwargs)

    def setup_schedule(self):
        self.merge_inplace(self.app.conf.CELERY_BEAT_SCHEDULE)
        self.install_default_entries(self.schedule)
        self.update_from_dict(self.store.load_entries())

        self.sync()

    def sync(self):
        self.store.save_entries(self.schedule)


class ClusterScheduler(Scheduler):
    """
    Celery cluster-scheduler.

    Only single instance of running beat instances will be active.
    Another instances will run in "sleep-mode" and will waiting
    when master instance will dead[or not :-)].
    """

    def __init__(self, app, *args, **kwargs):
        self.lock_ttl = getattr(app.conf, 'CELERY_BEAT_STORE_LOCK_TTL', 60)

        super().__init__(app, *args, **kwargs)

    def acquire_lock(self):
        acquired = self.store.acquire_lock(self.lock_ttl)

        if acquired:
            self.setup_schedule()

        return acquired

    def setup_schedule(self):
        if self.store.has_locked():
            super().setup_schedule()

    def tick(self, *args, **kwargs):
        if not self.store.has_locked():
            acquired = self.acquire_lock()

            if acquired:
                logger.info('beat: Lock acquired.')
            else:
                logger.info('beat: Awaiting lock...')
                return self.max_interval
        else:
            self.store.renew_lock()
            logger.debug('beat: Lock renewed.')

        return super().tick(*args, **kwargs)

    def sync(self):
        if not self.store.has_locked():
            return

        super().sync()

    def close(self):
        super().close()

        if self.store.has_locked():
            self.store.release_lock()
            logger.info('beat: Lock released.')



import json

from redis import StrictRedis

from beatx import serializer


class Store:
    SCHEDULE_KEY = 'celery:schedule'
    LOCK_KEY = 'celery:beat-lock'

    def __init__(self, store):
        self.rdb = StrictRedis.from_url(store)
        self.lock = None

    def load_entries(self):
        return {
            name.decode(): serializer.deserialize_entry(
                json.loads(data.decode())
            ) for name, data in self.rdb.hgetall(self.SCHEDULE_KEY).items()
        }

    def save_entries(self, entries):
        self.rdb.hmset(self.SCHEDULE_KEY, {
            name: json.dumps(
                serializer.serialize_entry(entry)
            ) for name, entry in entries.items()
        })

    def has_locked(self):
        return self.lock is not None

    def acquire_lock(self, lock_ttl=60):
        lock = self.rdb.lock(self.LOCK_KEY, timeout=lock_ttl, sleep=1)

        if lock.acquire(blocking=False):
            self.lock = lock
            return True

        return False

    def renew_lock(self):
        if not self.has_locked():
            raise RuntimeError("Cannot renew lock: lock is not set.")

        self.rdb.pexpire(self.LOCK_KEY, self.lock.timeout * 1000)

    def release_lock(self):
        self.lock.release()
        self.lock = None



from setuptools import setup, find_packages

setup(
    name='celery-beatx',
    version='0.3.0',
    url='https://github.com/mixkorshun/celery-beatx',
    description='Adds scheduler and cluster-scheduler for celery',
    keywords=['celery', 'celerybeat', 'scheduler'],

    long_description=open('README.rst', 'r').read(),

    author='Vladislav Bakin',
    author_email='mixkorshun@gmail.com',
    maintainer='Vladislav Bakin',
    maintainer_email='mixkorshun@gmail.com',

    license='MIT',
    setup_requires=[
        'pytest-runner',
    ],

    install_requires=[
        'celery',
    ],

    packages=find_packages(exclude=['tests.*', 'tests']),

    test_suite='tests',
    test_requires=['pytest', 'ephem', 'redis'],

    classifiers=[
        # 'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from datetime import datetime

from celery.beat import ScheduleEntry
from celery.schedules import crontab, schedule, solar

__all__ = ('serialize_entry', 'deserialize_entry')

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'


def encode_datetime(dt, allow_null=False):
    if dt is None and not allow_null:
        raise ValueError('Cannot encode datetime. None is not datetime.')

    return dt.strftime(DATETIME_FORMAT) if dt else None


def decode_datetime(s, allow_null=False):
    if s is None and not allow_null:
        raise ValueError('Cannot decode datetime. None is not datetime.')

    return datetime.strptime(s, DATETIME_FORMAT) if s else None


def encode_schedule(value):
    if value is None:
        return None
    elif isinstance(value, datetime):
        return {
            '__type__': 'datetime',
            '__value__': encode_datetime(value)
        }
    elif isinstance(value, crontab):
        return {
            '__type__': 'crontab',
            '__value__': '%(minute)s\t%(hour)s\t%(day_of_week)s\t'
                         '%(day_of_month)s\t%(month_of_year)s' % {
                             'minute': value._orig_minute,
                             'hour': value._orig_hour,
                             'day_of_week': value._orig_day_of_week,
                             'day_of_month': value._orig_day_of_month,
                             'month_of_year': value._orig_month_of_year,
                         }
        }
    elif isinstance(value, solar):
        return {
            '__type__': 'solar',
            '__value__': {
                'event': value.event,
                'lat': value.lat,
                'lon': value.lon
            }
        }
    elif isinstance(value, schedule):
        return {
            '__type__': 'schedule',
            '__value__': {
                'run_every': value.run_every.total_seconds(),
                'relative': bool(value.relative),
            }
        }
    else:
        raise NotImplementedError(
            'Cannot serialize schedule %(type)s type' % {
                'type': type(value).__name__
            }
        )


def decode_schedule(obj):
    if obj is None:
        return None

    _type = obj['__type__']
    value = obj['__value__']

    if _type == 'datetime':
        return decode_datetime(value)
    elif _type == 'crontab':
        return crontab(*value.split('\t'))
    elif _type == 'solar':
        return solar(**value)
    elif _type == 'schedule':
        return schedule(**value)
    else:
        raise NotImplementedError(
            'Cannot deserialize schedule %(type)s type' % {
                'type': _type
            }
        )


def serialize_entry(entry, schedule_encoder=encode_schedule):
    return {
        'name': entry.name,
        'task': entry.task,
        'schedule': schedule_encoder(entry.schedule),
        'args': entry.args,
        'kwargs': entry.kwargs,
        'last_run_at': encode_datetime(entry.last_run_at, allow_null=True),
        'total_run_count': entry.total_run_count,
        'options': entry.options
    }


def deserialize_entry(entry, schedule_decoder=decode_schedule):
    return ScheduleEntry(
        name=entry['name'],
        task=entry['task'],
        schedule=schedule_decoder(entry['schedule']),
        args=entry['args'],
        kwargs=entry['kwargs'],
        last_run_at=decode_datetime(entry['last_run_at'], allow_null=True),
        total_run_count=entry['total_run_count'],
        options=entry['options'],
    )



from importlib import import_module


def import_string(dotted_path):
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        raise ImportError("%s doesn't look like a module path" % dotted_path)

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        )
